/***************************************************
*                                                  *
*            C I R C _ B U F                       *
*            ===============                       *
*                                                  *
*  Comments: Circluar buffer routines for				*
*			    maintenamce,insertion into, and			*
*				 extraction from, status of,etc..		*
*																	*
*	Compilation Instructions:								*
*		Compile this file with TEST set to FALSE		*
*		to use with your routines. Set TEST to TRUE	*
*		to test these routines by themseleves.			*
*																	*
*	Linking Instructions:									*
*		Not a main file (unless you compile with		*
*		TEST set to TRUE).  Must be linked with		*
*		a main().												*
*																	*
*	References Used: 											*
*		Interrupts and the IBM PC Parts 1&2,			*
*		Nov/Dec 83,January 84,PC Tech Journal			*
*																	*		  
*												   				*
*  Programmer(s):  Al Morgan                       *
*																	*
*  Revision 1.0									   		*
*  Revision Date:09/17/1987								*
*																	*
*	References Used: 											*
*												   				*
***************************************************/


	#include <stdio.h>
	#include <dos.h>
	#include "circ_buf.h"			/* contains circular buffer definitions */
	#define	TRUE	1
	#define	FALSE	0

	#define	TEST		FALSE		/* Set to TRUE for testing the routines in this file */ 

#if TEST==TRUE
	circ_buffer_description buffer_descript;

	char test_buffer[]={"01234567890"};
	void test_cir(void);
	void reset_circ_buffer(circ_buffer_description *);

main()
{
	printf("Testing %s\n",__FILE__);
	printf("Version Date: %s, Time: %s\n",__DATE__,__TIME__);

	test_cir();
}
void test_cir(void)
{
	char in_string[100],tmp[100];
	int count,i;

	/* Test Case 1 */
	for(i=0;i<100;i++)
		in_string[i]='0'+i;
	for(i=0;i<sizeof(test_buffer);i++)
		test_buffer[i]='.';

	buffer_descript.head 	 = test_buffer;
	buffer_descript.tail 	 = test_buffer;
	buffer_descript.address  = test_buffer;
	buffer_descript.length		 = 10;
	buffer_descript.char_count	 = 0;

	for (i=0;i<10;i++)
	{
		printf("\nCirc_buf Test case %d\n",i);
	 	print_test_buffer(test_buffer);
		print_bufr_descript(&buffer_descript);
		count = puts_circ_buffer (in_string,3,&buffer_descript);
		if(count != 3)
			printf("Error on puts_circ_buffer()\n");
		if(check_circ_buffer_fill(&buffer_descript) != buffer_descript.char_count)
			printf("Error in check_cir_buffer_fill()\n");
		if (check_circ_buffer_over(&buffer_descript))
			printf("Error in check_cir_over()\n");
		count = gets_circ_buffer(tmp,1,&buffer_descript);
		if(count != 1)
			printf("Error on gets_circ_buffer()\n");
		print_bufr_descript(&buffer_descript);
	 	print_test_buffer(test_buffer);
	}
	puts("\nReset Circ buffer Test\n");
	reset_circ_buffer(&buffer_descript);
	print_bufr_descript(&buffer_descript);
}

print_test_buffer(char *string)
{
	int i;
	for(i=0;i<10;i++)
		printf("%c",string[i]);
	puts("");
}
print_bufr_descript(cb)
	circ_buffer_description *cb;
{
	printf("Head pointer    = %d\n",cb->head);
	printf("Tail pointer    = %d\n",cb->tail);
	printf("Address of buf  = %d\n",cb->address);
	printf("Length of buf   = %d\n",cb->length);
	printf("Character count = %d\n",cb->char_count);
	printf("Overflow flag   = %d\n",cb->overflow_flag);
}
	
#endif
/*****************************************
**                                      **
**     Gets_circ_buffer                 **
**									    			 **
*****************************************/
int gets_circ_buffer(target_buffer, bytes_wanted, cb)
	char *target_buffer;
	int bytes_wanted;
	circ_buffer_description *cb;
{
/*	
	This routine gets a string out of a circular buffer pointed to by
	cb of bytes_wanted size (if there are bytes_wanted characters in
	the circular buffer).  If no characters are in the buffer it returns
	a ZERO. If less characters are in the buffer than what is wanted 
	it only returns the characters in the buffer and indicates to the 
	caller the actual character count returned.
																							*/
	char *circ_buffer_end;
	int bytes_available,left_off;

	if(!cb->char_count)	/* if no characters in buffer,don't bother to proceed */
		return(0);

	circ_buffer_end = cb->address + cb->length; /* calc where end of circ buffer is */

	if (cb->head > cb->tail)	/* Is tail pointer "behind" head,ie. No buffer wrap? */
	{
		/* Case 1:	No buffer 'wrap' */
		#if TEST==TRUE
		puts("Processing a NoWrap condition");
		#endif
		bytes_available = cb->head-cb->tail;			/* number of bytes available is simply distance between head and tail */
		if(bytes_available<bytes_wanted)					/* Check to see if more bytes are being asked for than what we have */
			bytes_wanted = bytes_available; 				/* if so,reduce request to what we have */
		movmem(cb->tail,target_buffer,bytes_available);	/* move bytes_wanted to begining at tail,to target buffer */
		cb->tail = cb->tail + bytes_wanted;				/* update tail pointer by adding actual bytes given to user */
		disable();												/* inhibit interrupts while changing the char count,too avoid concurrent modification */
		cb->char_count-=bytes_wanted;						/* decrement the char count by number of chars taken out of buffer */
		cb->overflow_flag=0;									/* reset overflow flag,incase it was set */
		enable();												/* allow interrupts */
		return(bytes_wanted);								/* let caller know how many bytes where actually returned */
	}
	else
	{
		/* Case 2 ,Head pointer has wrapped around and behind tail pointer */
		#if TEST==TRUE
		puts("Processing a Wrap condition");
  		#endif
 
		/* First check if caller only wants the bytes at end of circ_buffer */
		/* IE. callers' request will not cause the tail pntr to wrap around */
		/* to begining of circ_buffer once we give the caller his bytes */

		bytes_available = circ_buffer_end - cb->tail;	/* For now say bytes avail is only to end of circ buffer */
		if(bytes_wanted <= bytes_available)	/* Check, if he wants only end of buffer bytes */
		{
			#if TEST==TRUE
			puts("..But no wrapping on extracting");
			#endif
			movmem(cb->tail,target_buffer,bytes_wanted); /* move bytes_wanted to begining at tail,to target buffer */
			cb->tail = cb->tail + bytes_wanted;				/* update tail pointer by adding actual bytes given to user */
			disable();												/* inhibit interrupts while changing the char count,too avoid concurrent modification */
			cb->char_count-=bytes_wanted;						/* decrement the char count by number of chars taken out of buffer */
			cb->overflow_flag=0;									/* reset overflow flag,incase it was set */
			enable();												/* allow interrupts */
			return(bytes_wanted);								/* let caller know how many bytes where actually returned */
		}
		else
		{
			#if TEST==TRUE
				puts("...with extraction wrapping");
			#endif
		/* Nope, Caller wants so many bytes it will wrap past the circ_buffer_end */
		/* and suck up some bytes at the circ_buffer_beginning			*/

		/* Num of bytes available is the end of buffer bytes (tail-circ_buffer_end) */
		/* plus the begin of buffer bytes(cb->head-cb->address)			*/

			bytes_available =(circ_buffer_end-cb->tail)+(cb->head-cb->address); 
			if(bytes_available<bytes_wanted) 				/* if caller wants more than we have.. */
				bytes_wanted=bytes_available;					/* only give the caller what we have */
			movmem(cb->tail,target_buffer,circ_buffer_end-cb->tail); /* move the bytes at circ_buffer's end */
			left_off = circ_buffer_end-cb->tail;			/* calc what we just gave him (so far) */
			cb->tail = cb->address;								/* temporarily say tail is the circ_buufer start */
			movmem(cb->tail,target_buffer+left_off,bytes_wanted-left_off); /* give the caller the bytes at the begin of buffer */
			cb->tail = cb->tail + (bytes_wanted-left_off); /* adust tail pntr by using what we gave him the second time */
			disable();												/* inhibit interrupts while changing the char count,too avoid concurrent modification */
			cb->char_count-=bytes_wanted;						/* decrement the char count by number of chars taken out of buffer */
			cb->overflow_flag=0;									/* reset overflow flag,incase it was set */
			enable();												/* allow interrupts */
			return(bytes_wanted);								/* Tell caller what we actually gave him */
		}
	}
}
/*****************************************
**                                      **
**     Puts_circ_buffer                 **
**									    			 **
*****************************************/
int puts_circ_buffer(source_buffer, bytes_to_move, cb)
	char *source_buffer;
	int bytes_to_move;
	circ_buffer_description *cb;
{
/*
	This routine accepts a string of characters (bytes_to_move) and
	places them into the buffer pointed to by cb.  If the buffer is not
	big enough the routine will not accept the string and return ZERO
	to indicate to the caller this has happened.  Otherwise it will accept
	the string and return the number of characters accepted.
																							*/
	char *circ_buffer_end;
	int bytes_available,left_off;

	if(cb->char_count+bytes_to_move > cb->length)	/* if not enough room left in buffer,don't bother to proceed */
		return(0);
	circ_buffer_end = cb->address + cb->length; /* calc where end of circ buffer is */

	if (cb->head >= cb->tail)	/* Is tail pointer "behind" head,ie. buffer wrap? */
	{
		#if TEST==TRUE
		puts("Processing a Potential Wrap condition");
		#endif

		/* First check if caller only wants to move bytes into the end of circ_buffer */
		/* IE. callers' request will not cause the head pntr to wrap around */
		/* to begining of circ_buffer once we get the caller's bytes */

		bytes_available = circ_buffer_end - cb->head;	/* For now say bytes avail is only to end of circ buffer */
		if(bytes_to_move <= bytes_available)				/* Check, if he wants to put at end of buffer */
		{
			#if TEST==TRUE
			puts("..But no wrapping on insertion");
			#endif
			movmem(source_buffer,cb->head,bytes_to_move); /* move bytes_to_move to begining at tail,to target buffer */
			cb->head = cb->head + bytes_to_move;			/* update tail pointer by adding actual bytes given to user */
			disable();												/* inhibit interrupts while changing the char count,too avoid concurrent modification */
			cb->char_count+=bytes_to_move;					/* update character count */
			enable();												/* allow interrupts */
			return(bytes_to_move);								/* let caller know how many bytes where actually returned */
		}
		else
		{
			#if TEST==TRUE
				puts("...with insertion wrapping");
			#endif
		/* Nope, Caller wants to put in so many bytes it will wrap past */
		/* the circ_buffer_end and use up some bytes at the */
		/* circ_buffer_beginning	*/

		/* Num of bytes available is the end of buffer bytes (head-circ_buffer_end) */
		/* plus the begin of buffer bytes(cb->tail-cb->address)			*/

			bytes_available =(circ_buffer_end-cb->head) + (cb->tail-cb->address); 
			movmem(source_buffer,cb->head,circ_buffer_end-cb->head); /* move the bytes at circ_buffer's end */
			left_off = circ_buffer_end-cb->head;			/* calc what we just gave him (so far) */
			cb->head = cb->address;				  				/* temporarily say head is the circ_buufer start */
			movmem(source_buffer+left_off,cb->head,bytes_to_move-left_off); /* give the caller the bytes at the begin of buffer */
			cb->head = cb->head + (bytes_to_move-left_off); /* adust head pntr by using what we gave him the second time */
			disable();												/* inhibit interrupts while changing the char count,too avoid concurrent modification */
			cb->char_count+=bytes_to_move;					/* update character count */
			enable();												/* allow interrupts */
			return(bytes_to_move);								/* Tell caller what we actually gave him */
		}
	}
	else
	{
		/* Case 1:	No buffer 'wrap' */
		#if TEST==TRUE
		puts("Processing a NoWrap condition");
		#endif
		movmem(source_buffer,cb->head,bytes_to_move);	/* move bytes_to_move to begining at head,to target buffer */
		cb->head = cb->head + bytes_to_move;				/* update head pointer by adding actual bytes given to user */
		disable();													/* inhibit interrupts while changing the char count,too avoid concurrent modification */
		cb->char_count+=bytes_to_move;						/* update character count */
		enable();													/* allow interrupts */
		return(bytes_to_move);									/* let caller know how many bytes where actually returned */
	 }
}
/*****************************************
**                                      **
**     Reset_circ_buffer                **
**									    			 **
*****************************************/
void reset_circ_buffer(cb)
	circ_buffer_description *cb;
{
	disable();						/* disable interrupts, to avoid concurrent mods */
	cb->head 	 = cb->address;
	cb->tail 	 = cb->head;
	cb->char_count	 = 0;
	cb->overflow_flag = 0;
	enable();						/* allow interrupts to occur */
}
/*****************************************
**                                      **
**     Check_circ_buffer_fill           **
**									    			 **
*****************************************/
int check_circ_buffer_fill(cb)
	circ_buffer_description *cb;
{
/* 
	This routine returns the number of characters in the circular
	buffer pointed to by the passed circular buffer description.

																							*/
	return(cb->char_count);
}
/*****************************************
**                                      **
**     Check_circ_buffer_over           **
**									    			 **
*****************************************/
int check_circ_buffer_over(cb)
	circ_buffer_description *cb;
{
/* 
	This routine returns the overflow flag for the circular
	buffer pointed to by the passed circular buffer description.

																							*/
	return(cb->overflow_flag);
}


